#include "Centipede.h"
#include "CentipedeSegment.h"
#include "Definitions.h"
#include "CentipedeHead.h"
#include "CentipedeBody.h"
#include "CentipedeManager.h"
#include "MushroomFieldManager.h"
#include "WaveManager.h"
#include "Wave.h"
#include "MoveState.h"

Centipede::Centipede() {
	splitRequested = false;
	headKillRequested = false;
}

Centipede::~Centipede() {
	CentipedeNode* pCurrentNode = this->pCentipedeHead;
	while (pCurrentNode != nullptr) {
		CentipedeNode* tmp = pCurrentNode->pNext;
		delete pCurrentNode;
		pCurrentNode = tmp;
	}
}

void Centipede::SplitUpdate() {
	if (splitRequested) {
		if (!splitRequestedAt.empty()) {
			//so essentially, we need to make sure the list is in backwards order from back to front.

			//first, starting from the back of the segment's list, we check if the item is in our centipede.

			CentipedeNode* pCurrentNode = this->pCentipedeEnd;
			while (pCurrentNode != nullptr) {
				CentipedeNode* pNextNode = pCurrentNode->pPrev;
				for (std::vector<CentipedeSegment*>::iterator it = splitRequestedAt.begin(); it != splitRequestedAt.end(); ++it) {
					//check each element in the split list to see if it's in the list.
					if ((*it) == pCurrentNode->pSeg) {
						this->Split(*it);
						splitRequestedAt.erase(it);
						break;
					}
				}
				pCurrentNode = pNextNode;
			}
		}
		splitRequested = false;
	}
	else if (headKillRequested) {
		this->KillHead();
		headKillRequested = false;
	}
}

void Centipede::Update() {
	//this method is called every 8 (cn::CENTIPEDE FRAMES) frames. The idea is that when called, the centipede moves its movement down the chain from here.
	//it calls from the back of the list in order to preserve the correct offset data
	CentipedeNode* pCurrentNode = this->pCentipedeEnd;
	while (pCurrentNode->pPrev != nullptr) {
		pCurrentNode->pSeg->nextGridFrame();
		pCurrentNode->pSeg->InheritMovement(pCurrentNode->pPrev->pSeg);
		pCurrentNode = pCurrentNode->pPrev;
	}
	this->pCentipedeHead->pSeg->nextGridFrame();

}

void Centipede::Initialize(int spd, int length, bool rnd) {
	int grid_x = cn::MUSHROOM_FIELD_COLUMNS / 2;
	if (rnd) {
		grid_x = rand() % cn::MUSHROOM_FIELD_COLUMNS;
	}
	this->pCentipedeHead = new CentipedeNode;
	this->pCentipedeHead->pSeg = CentipedeManager::RequestHead(this, spd, grid_x);
	//Trace::out("Registered pHead: %p to Centipede: %p\n", this->pCentipedeHead->pSeg, this);
	CentipedeNode* pCurrentNode = this->pCentipedeHead;
	if (length > 1) {
		for (int i = 0; i < (length - 1); i++) {
			pCurrentNode->pNext = new CentipedeNode();
			pCurrentNode->pNext->pSeg = CentipedeManager::RequestBody(this, spd, grid_x);
			pCurrentNode->pNext->pPrev = pCurrentNode;
			pCurrentNode = pCurrentNode->pNext;
		}
	}
	this->pCentipedeEnd = pCurrentNode;
}

void Centipede::Deinitialize() {
	CentipedeNode* pCurrentNode = this->pCentipedeHead;
	while (pCurrentNode != nullptr) {
		CentipedeNode* tmp = pCurrentNode->pNext;
		delete pCurrentNode;
		pCurrentNode = tmp;
	}
	this->pCentipedeHead = nullptr;
	this->pCentipedeEnd = nullptr;
}

void Centipede::InitializeWithData(CentipedeNode* pHead, CentipedeNode* pEnd) {
	this->pCentipedeHead = pHead;
	this->pCentipedeEnd = pEnd;
}

CentipedeNode* Centipede::findNode(CentipedeSegment* pSeg) {
	CentipedeNode* pCurrent = this->pCentipedeHead;
	while (pCurrent != nullptr) {
		if (pCurrent->pSeg == pSeg)
			return pCurrent;
		pCurrent = pCurrent->pNext;
	}
	//does not exist
	assert(0);
	return nullptr;
}

void Centipede::Split(CentipedeSegment* pSeg) {
	sf::Vector2<char> copy = *pSeg->getMoveState()->getGridOffsets();
	MushroomFieldManager::AddMushroomToCurrentField(pSeg->getGridPos()+copy);
	//pSeg->DeInitialize();
	CentipedeNode* pSegNode = findNode(pSeg);
	if (pSegNode == this->pCentipedeEnd) {
		//Trace::out("Shot the end of the centipede");
		this->pCentipedeEnd = this->pCentipedeEnd->pPrev;
		this->pCentipedeEnd->pNext = nullptr;
	}
	else {
		CentipedeNode* endHolder = this->pCentipedeEnd;
		this->pCentipedeEnd = pSegNode->pPrev;
		Centipede* nCentipede = CentipedeManager::GetNewCentipede_toInsert();
		nCentipede->InitializeWithData(pSegNode->pNext, endHolder);
		CentipedeHead* tHead = CentipedeManager::UpgradeToHead(static_cast<CentipedeBody*>(nCentipede->pCentipedeHead->pSeg));
		//nCentipede->pCentipedeHead->pSeg->DeInitialize();
		nCentipede->pCentipedeHead->pSeg->MarkForDestroy();
		nCentipede->pCentipedeHead->pSeg = tHead;
		this->pCentipedeEnd->pNext = nullptr;
		nCentipede->pCentipedeHead->pPrev = nullptr;
		
		CentipedeNode* nIt = nCentipede->pCentipedeHead;
		while (nIt != nullptr) {
			nIt->pSeg->setMother(nCentipede);
			nIt = nIt->pNext;
		}
		//nCentipede->pCentipedeHead->pSeg->assureGridFrame();
	}
	delete pSegNode;
	pSeg->MarkForDestroy();
	

}

void Centipede::KillHead() {
	sf::Vector2<char> copy = *pCentipedeHead->pSeg->getMoveState()->getGridOffsets();
	MushroomFieldManager::AddMushroomToCurrentField(pCentipedeHead->pSeg->getGridPos()+copy);
	this->pCentipedeHead->pSeg->MarkForDestroy();
	if (this->pCentipedeHead->pNext == nullptr) {
		CentipedeManager::ToBeRecycled(this);
		//delete this->pCentipedeHead;
 		return;
	}
	CentipedeNode* toBeDestroyed = this->pCentipedeHead;
	this->pCentipedeHead = this->pCentipedeHead->pNext;
	CentipedeHead* tHead = CentipedeManager::UpgradeToHead(static_cast<CentipedeBody*>(this->pCentipedeHead->pSeg));
	this->pCentipedeHead->pSeg->MarkForDestroy();
	this->pCentipedeHead->pSeg = tHead;
	this->pCentipedeHead->pPrev = nullptr;
	delete toBeDestroyed;
	//this->pCentipedeHead->pSeg->assureGridFrame();

}

void Centipede::KillAll() {
	//so the segment data needs to be deleted here, however, the node data will be deinitialized later, so no need to delete.

	CentipedeNode* nIt = this->pCentipedeHead;
	while (nIt != nullptr) {
		nIt->pSeg->MarkForDestroy();
		nIt = nIt->pNext;
	}
	splitRequested = false;
	headKillRequested = false;
	CentipedeManager::MarkForImmediateRecycle(this);
}
